from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db import IntegrityError
from django.db.models import Max
from django.forms.forms import ValidationError
from django.utils.translation import gettext as _

import graphene
from graphene_django.types import DjangoObjectType
from graphql import GraphQLError

from aleksis.apps.chronos.models import Subject
from aleksis.core.models import Group, Room
from aleksis.core.schema.base import DeleteMutation
from aleksis.core.schema.room import RoomType

from .models import Seat, SeatingPlan
from .util.perms import get_allowed_seating_plans


class SeatType(DjangoObjectType):
    class Meta:
        model = Seat
        fields = ("id", "x", "y", "person", "seated", "plan")

    @staticmethod
    def resolve_person(root, info):
        if not info.context.user.has_perm("core.view_person_rule", root.person):
            raise PermissionDenied()
        return root.person

    @staticmethod
    def resolve_plan(root, info):
        if not info.context.user.has_perm("stoelindeling.view_seatingplan_rule", root.plan):
            raise PermissionDenied()
        return root.plan


class SeatingPlanType(DjangoObjectType):
    class Meta:
        model = SeatingPlan
        fields = "__all__"

    name = graphene.String()
    width = graphene.Int()
    height = graphene.Int()

    can_edit = graphene.Boolean()
    can_delete = graphene.Boolean()

    room = graphene.Field(RoomType)

    def resolve_name(root, info):  # noqa
        return str(root)

    def resolve_width(root, info):  # noqa
        return root.seats.aggregate(Max("x"))["x__max"]

    def resolve_height(root, info):  # noqa
        return root.seats.aggregate(Max("y"))["y__max"]

    @staticmethod
    def resolve_can_edit(root, info):
        return info.context.user.has_perm("stoelindeling.edit_seatingplan_rule", root)

    @staticmethod
    def resolve_can_delete(root, info):
        return info.context.user.has_perm("stoelindeling.delete_seatingplan_rule", root)

    @staticmethod
    def resolve_seats(root, info):
        return root.get_all_seats()


class SubjectType(DjangoObjectType):
    """
    GraphQL Type for Chronos subjects.

    This functions as a leaf node in the graph (because only attributes of
    primitive types are accessible). Therefore, no permission checks are
    necessary inside this type.
    """

    class Meta:
        model = Subject
        fields = ("id", "name", "short_name")


class ChoiceType(graphene.ObjectType):
    """
    Base Type for choices in Stoelindeling.

    Used to created subclasses for choices for Groups, Rooms and Subjects,
    as object of different model types with same IDs would have the same
    GraphQL type and would therefore be considered equal by many GraphQL
    clients which could lead to unwanted side effects.
    """

    id = graphene.ID()  # noqa
    name = graphene.String()

    @staticmethod
    def resolve_name(root, info):
        return str(root)


class GroupChoiceType(ChoiceType):
    ...


class RoomChoiceType(ChoiceType):
    ...


class SubjectChoiceType(ChoiceType):
    ...


class SeatingPlanChoicesType(graphene.ObjectType):
    groups = graphene.List(GroupChoiceType)
    subjects = graphene.List(SubjectChoiceType)
    rooms = graphene.List(RoomChoiceType)

    @staticmethod
    def resolve_groups(root, info):
        qs = Group.objects.all()
        if not info.context.user.has_perm("stoelindeling.create_seatingplan_rule"):
            qs = qs.filter(owners=info.context.user.person)
        return qs

    @staticmethod
    def resolve_subjects(root, info):
        if not info.context.user.has_perm("stoelindeling.create_seatingplan_rule"):
            return []
        return Subject.objects.all()

    @staticmethod
    def resolve_rooms(root, info):
        if not info.context.user.has_perm("stoelindeling.create_seatingplan_rule"):
            return []
        return Room.objects.all()


class Query(graphene.ObjectType):
    seating_plan_by_id = graphene.Field(SeatingPlanType, id=graphene.ID())
    all_seating_plans = graphene.List(SeatingPlanType)
    seating_plan_choices = graphene.Field(SeatingPlanChoicesType, resolver=SeatingPlanChoicesType)

    def resolve_seating_plan_by_id(root, info, id):  # noqa
        plan = SeatingPlan.objects.get(id=id)
        if not info.context.user.has_perm("stoelindeling.view_seatingplan_rule", plan):
            raise PermissionDenied
        return plan

    def resolve_all_seating_plans(root, info):  # noqa
        return get_allowed_seating_plans(info.context.user)


class MoveSeatMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID()  # noqa
        x = graphene.Int()
        y = graphene.Int()
        seated = graphene.Boolean()

    seat = graphene.Field(SeatType)

    @classmethod
    def mutate(cls, root, info, id, x, y, seated):  # noqa
        seat = Seat.objects.get(id=id)
        seat.x = x
        seat.y = y
        seat.seated = seated
        seat.save()
        return MoveSeatMutation(seat=seat)


class CreateSeatingPlanMutation(graphene.Mutation):
    class Arguments:
        room = graphene.ID()
        group = graphene.ID()
        subject = graphene.ID(required=False)

    seating_plan = graphene.Field(SeatingPlanType)

    @classmethod
    def mutate(cls, root, info, room, group, subject=None):
        if not info.context.user.has_perm("stoelindeling.create_seatingplan_rule"):
            raise PermissionDenied

        group_object = Group.objects.get(id=group)
        room_object = Room.objects.get(id=room)
        subject_object = Subject.objects.get(id=subject) if subject is not None else None

        if group_object.subject and subject_object and group_object.subject != subject_object:
            raise ValidationError(
                _(
                    "The group you selected has a fixed subject ({}). "
                    "You are not allowed to use another subject than it."
                ).format(group_object.subject)
            )

        plan = SeatingPlan.objects.create(
            room=room_object, group=group_object, subject=subject_object
        )
        plan.save()

        messages.success(info.context, _("The seating plan has been created."))

        return CreateSeatingPlanMutation(seating_plan=plan)


class EditSeatingPlanMutation(graphene.Mutation):
    class Arguments:
        plan_id = graphene.ID()
        room = graphene.ID()
        subject = graphene.ID(required=False)

    seating_plan = graphene.Field(SeatingPlanType)

    @classmethod
    def mutate(cls, root, info, plan_id, room, subject=None):
        try:
            plan = SeatingPlan.objects.get(id=plan_id)
            room = Room.objects.get(id=room)
            if subject is not None:
                subject = Subject.objects.get(id=subject)

            if not info.context.user.has_perm("stoelindeling.edit_seatingplan_rule", plan):
                raise PermissionDenied
        except ObjectDoesNotExist as e:
            raise GraphQLError(
                message="Query matching object does not exist",
                original_error=e,
            )

        plan.room = room
        plan.subject = subject

        try:
            plan.save()
        except IntegrityError as e:
            print("The error gets raised:")
            raise GraphQLError(
                message="Error: The Room and Subject values violate the unique_constraint.",
                original_error=e,
            )

        return EditSeatingPlanMutation(seating_plan=plan)


class DeleteSeatingPlanMutation(DeleteMutation):
    klass = SeatingPlan
    permission_required = "stoelindeling.delete_seatingplan_rule"


class Mutation(graphene.ObjectType):
    move_seat = MoveSeatMutation.Field()
    create_seating_plan = CreateSeatingPlanMutation.Field()
    edit_seating_plan = EditSeatingPlanMutation.Field()
    delete_seating_plan = DeleteSeatingPlanMutation.Field()
